from typing import (
    AbstractSet,
    Any,
    Callable,
    ClassVar,
    Dict,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

from pydantic import BaseModel
from sqlalchemy.orm import (
    RelationshipProperty,
    declared_attr,
)
from sqlalchemy.orm.attributes import set_attribute
from sqlalchemy.orm.instrumentation import is_instrumented
from sqlalchemy.sql.schema import MetaData
from sqlmodel.main import (
    IS_PYDANTIC_V2,
    PYDANTIC_VERSION,
    SQLModelConfig,
    _calculate_keys,
    finish_init,
    get_model_fields,
    init_pydantic_private_attrs,
    is_table_model_class,
    sqlmodel_init,
    sqlmodel_validate,
    default_registry,
    _TSQLModel,
    IncEx,
)
from typing_extensions import Literal, deprecated
from autoflow.utils.sql_model_metaclass import PatchSQLModelMetaclass


class PatchSQLModel(
    BaseModel, metaclass=PatchSQLModelMetaclass, registry=default_registry
):
    # SQLAlchemy needs to set weakref(s), Pydantic will set the other slots values
    __slots__ = ("__weakref__",)
    __tablename__: ClassVar[Union[str, Callable[..., str]]]
    __sqlmodel_relationships__: ClassVar[Dict[str, RelationshipProperty[Any]]]
    __name__: ClassVar[str]
    metadata: ClassVar[MetaData]
    __allow_unmapped__ = True  # https://docs.sqlalchemy.org/en/20/changelog/migration_20.html#migration-20-step-six

    if IS_PYDANTIC_V2:
        model_config = SQLModelConfig(from_attributes=True)
    else:

        class Config:
            orm_mode = True

    def __new__(cls, *args: Any, **kwargs: Any) -> Any:
        new_object = super().__new__(cls)
        # SQLAlchemy doesn't call __init__ on the base class when querying from DB
        # Ref: https://docs.sqlalchemy.org/en/14/orm/constructors.html
        # Set __fields_set__ here, that would have been set when calling __init__
        # in the Pydantic model so that when SQLAlchemy sets attributes that are
        # added (e.g. when querying from DB) to the __fields_set__, this already exists
        init_pydantic_private_attrs(new_object)
        return new_object

    def __init__(__pydantic_self__, **data: Any) -> None:
        # Uses something other than `self` the first arg to allow "self" as a
        # settable attribute

        # SQLAlchemy does very dark black magic and modifies the __init__ method in
        # sqlalchemy.orm.instrumentation._generate_init()
        # so, to make SQLAlchemy work, it's needed to explicitly call __init__ to
        # trigger all the SQLAlchemy logic, it doesn't work using cls.__new__, setting
        # attributes obj.__dict__, etc. The __init__ method has to be called. But
        # there are cases where calling all the default logic is not ideal, e.g.
        # when calling Model.model_validate(), as the validation is done outside
        # of instance creation.
        # At the same time, __init__ is what users would normally call, by creating
        # a new instance, which should have validation and all the default logic.
        # So, to be able to set up the internal SQLAlchemy logic alone without
        # executing the rest, and support things like Model.model_validate(), we
        # use a contextvar to know if we should execute everything.
        if finish_init.get():
            sqlmodel_init(self=__pydantic_self__, data=data)

    def __setattr__(self, name: str, value: Any) -> None:
        if name in {"_sa_instance_state"}:
            self.__dict__[name] = value
            return
        else:
            # Set in SQLAlchemy, before Pydantic to trigger events and updates
            if is_table_model_class(self.__class__) and is_instrumented(self, name):  # type: ignore[no-untyped-call]
                set_attribute(self, name, value)
            # Set in Pydantic model to trigger possible validation changes, only for
            # non relationship values
            if name not in self.__sqlmodel_relationships__:
                super().__setattr__(name, value)

    def __repr_args__(self) -> Sequence[Tuple[Optional[str], Any]]:
        # Don't show SQLAlchemy private attributes
        return [
            (k, v)
            for k, v in super().__repr_args__()
            if not (isinstance(k, str) and k.startswith("_sa_"))
        ]

    @declared_attr  # type: ignore
    def __tablename__(cls) -> str:
        return cls.__name__.lower()

    @classmethod
    def model_validate(
        cls: Type[_TSQLModel],
        obj: Any,
        *,
        strict: Union[bool, None] = None,
        from_attributes: Union[bool, None] = None,
        context: Union[Dict[str, Any], None] = None,
        update: Union[Dict[str, Any], None] = None,
    ) -> _TSQLModel:
        return sqlmodel_validate(
            cls=cls,
            obj=obj,
            strict=strict,
            from_attributes=from_attributes,
            context=context,
            update=update,
        )

    def model_dump(
        self,
        *,
        mode: Union[Literal["json", "python"], str] = "python",
        include: IncEx = None,
        exclude: IncEx = None,
        context: Union[Dict[str, Any], None] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        round_trip: bool = False,
        warnings: Union[bool, Literal["none", "warn", "error"]] = True,
        serialize_as_any: bool = False,
    ) -> Dict[str, Any]:
        if PYDANTIC_VERSION >= "2.7.0":
            extra_kwargs: Dict[str, Any] = {
                "context": context,
                "serialize_as_any": serialize_as_any,
            }
        else:
            extra_kwargs = {}
        if IS_PYDANTIC_V2:
            return super().model_dump(
                mode=mode,
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                round_trip=round_trip,
                warnings=warnings,
                **extra_kwargs,
            )
        else:
            return super().dict(
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
            )

    @deprecated(
        """
        🚨 `obj.dict()` was deprecated in SQLModel 0.0.14, you should
        instead use `obj.model_dump()`.
        """
    )
    def dict(
        self,
        *,
        include: IncEx = None,
        exclude: IncEx = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> Dict[str, Any]:
        return self.model_dump(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )

    @classmethod
    @deprecated(
        """
        🚨 `obj.from_orm(data)` was deprecated in SQLModel 0.0.14, you should
        instead use `obj.model_validate(data)`.
        """
    )
    def from_orm(
        cls: Type[_TSQLModel], obj: Any, update: Optional[Dict[str, Any]] = None
    ) -> _TSQLModel:
        return cls.model_validate(obj, update=update)

    @classmethod
    @deprecated(
        """
        🚨 `obj.parse_obj(data)` was deprecated in SQLModel 0.0.14, you should
        instead use `obj.model_validate(data)`.
        """
    )
    def parse_obj(
        cls: Type[_TSQLModel], obj: Any, update: Optional[Dict[str, Any]] = None
    ) -> _TSQLModel:
        if not IS_PYDANTIC_V2:
            obj = cls._enforce_dict_if_root(obj)  # type: ignore[attr-defined] # noqa
        return cls.model_validate(obj, update=update)

    # From Pydantic, override to only show keys from fields, omit SQLAlchemy attributes
    @deprecated(
        """
        🚨 You should not access `obj._calculate_keys()` directly.

        It is only useful for Pydantic v1.X, you should probably upgrade to
        Pydantic v2.X.
        """,
        category=None,
    )
    def _calculate_keys(
        self,
        include: Optional[Mapping[Union[int, str], Any]],
        exclude: Optional[Mapping[Union[int, str], Any]],
        exclude_unset: bool,
        update: Optional[Dict[str, Any]] = None,
    ) -> Optional[AbstractSet[str]]:
        return _calculate_keys(
            self,
            include=include,
            exclude=exclude,
            exclude_unset=exclude_unset,
            update=update,
        )

    def sqlmodel_update(
        self: _TSQLModel,
        obj: Union[Dict[str, Any], BaseModel],
        *,
        update: Union[Dict[str, Any], None] = None,
    ) -> _TSQLModel:
        use_update = (update or {}).copy()
        if isinstance(obj, dict):
            for key, value in {**obj, **use_update}.items():
                if key in get_model_fields(self):
                    setattr(self, key, value)
        elif isinstance(obj, BaseModel):
            for key in get_model_fields(obj):
                if key in use_update:
                    value = use_update.pop(key)
                else:
                    value = getattr(obj, key)
                setattr(self, key, value)
            for remaining_key in use_update:
                if remaining_key in get_model_fields(self):
                    value = use_update.pop(remaining_key)
                    setattr(self, remaining_key, value)
        else:
            raise ValueError(
                "Can't use sqlmodel_update() with something that "
                f"is not a dict or SQLModel or Pydantic model: {obj}"
            )
        return self



import dspy
import logging
from typing import Any, List

import llama_index.core.instrumentation as instrument
from sqlmodel import SQLModel
from autoflow.indices.knowledge_graph.extractor import KnowledgeGraphExtractor
from autoflow.storage import KnowledgeGraphStore
from autoflow.schema import BaseComponent

logger = logging.getLogger(__name__)

dispatcher = instrument.get_dispatcher(__name__)


class KnowledgeGraphIndex(BaseComponent):
    def __init__(
        self,
        dspy_lm: dspy.LM,
        kg_store: KnowledgeGraphStore,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self._dspy_lm = dspy_lm
        self._kg_store = kg_store
        self._graph_extractor = KnowledgeGraphExtractor(dspy_lm=self._dspy_lm)

    def build_index_for_chunks(self, chunks: List[SQLModel]):
        if len(chunks) == 0:
            return chunks

        # TODO: make it faster.
        for chunk in chunks:
            pred = self._graph_extractor.forward(chunk.text)
            self._kg_store.save_knowledge_graph(pred.knowledge, chunk)



from enum import Enum

from llama_index.core.schema import BaseComponent


class DataSourceType(str, Enum):
    FILE = "file"
    WEB_SITEMAP = "web_sitemap"
    WEB_SINGLE_PAGE = "web_single_page"


class IndexMethod(str, Enum):
    VECTOR_SEARCH = "VECTOR_SEARCH"
    FULLTEXT_SEARCH = "FULLTEXT_SEARCH"
    KNOWLEDGE_GRAPH = "KNOWLEDGE_GRAPH"


BaseComponent = BaseComponent



from typing import Optional, List

from autoflow.llms.embeddings.litellm import LiteLLMEmbedding
from autoflow.storage.tidb.embed_fn import BaseEmbeddingFunction

EmbeddingModel = LiteLLMEmbedding


class EmbeddingFunction(BaseEmbeddingFunction):
    def __init__(self, model_name: str, dimensions: Optional[int] = None, **kwargs):
        self.embedding_model = LiteLLMEmbedding(
            model_name=model_name, dimensions=dimensions, **kwargs
        )
        if dimensions is not None:
            self._dimensions = dimensions
        else:
            self._dimensions = len(self.get_query_embedding("test"))

    def _get_dimensions(self) -> int:
        return self._dimensions

    def get_query_embedding(self, query: str) -> list[float]:
        return self.embedding_model.get_query_embedding(query)

    def get_source_embedding(self, source: str) -> list[float]:
        return self.embedding_model.get_text_embedding(source)

    def get_source_embedding_batch(self, sources: List[str]) -> list[list[float]]:
        return self.embedding_model.get_text_embedding_batch(texts=sources)


__all__ = ["EmbeddingModel", "EmbeddingFunction"]



from datetime import datetime
from typing import Optional, List, Dict, Type, Any
from uuid import UUID
from sqlalchemy import Column, Text, JSON, DateTime
from sqlalchemy.orm.decl_api import RegistryType
from sqlmodel import (
    SQLModel,
    Field,
    Relationship as SQLRelationship,
)
from sqlmodel.main import default_registry
from tidb_vector.sqlalchemy import VectorType

from autoflow.utils.sql_model import PatchSQLModel


def get_relationship_model(
    table_name: str,
    entity_db_model: Type[SQLModel],
    vector_dimension: int,
    registry: RegistryType = default_registry,
) -> Type[SQLModel]:
    entity_table_name = entity_db_model.__tablename__

    class Relationship(PatchSQLModel, table=True, registry=registry):
        __tablename__ = table_name
        __table_args__ = ({"extend_existing": True},)

        id: Optional[int] = Field(default=None, primary_key=True)

        description: str = Field(sa_column=Column(Text))
        description_vec: list[float] = Field(
            sa_column=Column(VectorType(vector_dimension))
        )
        source_entity_id: int = Field(foreign_key=f"{entity_table_name}.id")
        source_entity: entity_db_model = SQLRelationship(
            sa_relationship_kwargs={
                "primaryjoin": "Relationship.source_entity_id == Entity.id",
                "lazy": "joined",
            },
        )
        target_entity_id: int = Field(foreign_key=f"{entity_table_name}.id")
        target_entity: entity_db_model = SQLRelationship(
            sa_relationship_kwargs={
                "primaryjoin": "Relationship.target_entity_id == Entity.id",
                "lazy": "joined",
            },
        )

        meta: List | Dict = Field(default={}, sa_column=Column(JSON))
        weight: int = 0
        chunk_id: Optional[UUID] = Field(default=None, nullable=True)
        document_id: Optional[int] = Field(default=None, nullable=True)

        created_at: Optional[datetime] = Field(sa_column=Column(DateTime))
        updated_at: Optional[datetime] = Field(sa_column=Column(DateTime))

        def __hash__(self):
            return hash(self.id)

        def __eq__(self, other: Any) -> bool:
            return self.id == other.id

        def screenshot(self):
            obj_dict = self.model_dump(
                exclude={
                    "description_vec",
                    "source_entity",
                    "target_entity",
                }
            )
            return obj_dict

    return Relationship



import os
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import padding


class AESCipher:
    def __init__(self, key: bytes) -> None:
        self.key = key
        self.backend = default_backend()

    def encrypt(self, plain_text: str) -> bytes:
        # generate a random initialization vector
        iv = os.urandom(16)
        cipher = Cipher(algorithms.AES(self.key), modes.CFB(iv), backend=self.backend)
        encryptor = cipher.encryptor()

        # fill the last block with padding
        padder = padding.PKCS7(algorithms.AES.block_size).padder()
        padded_data = padder.update(plain_text.encode()) + padder.finalize()
        encrypted = encryptor.update(padded_data) + encryptor.finalize()
        return iv + encrypted

    def decrypt(self, encrypted_text: bytes) -> str:
        # get the initialization vector and the encrypted data
        iv = encrypted_text[:16]
        encrypted_data = encrypted_text[16:]

        cipher = Cipher(algorithms.AES(self.key), modes.CFB(iv), backend=self.backend)
        decryptor = cipher.decryptor()

        # remove the padding
        decrypted_padded = decryptor.update(encrypted_data) + decryptor.finalize()
        unpadder = padding.PKCS7(algorithms.AES.block_size).unpadder()
        decrypted = unpadder.update(decrypted_padded) + unpadder.finalize()
        return decrypted



from .base import DataSource
from .file import FileDataSource
from .web_sitemap import WebSitemapDataSource
from .web_single_page import WebSinglePageDataSource

__all__ = [
    "DataSource",
    "FileDataSource",
    "WebSitemapDataSource",
    "WebSinglePageDataSource",
]



import logging
from abc import ABC, abstractmethod
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Union,
    Tuple,
    Sequence,
)

import numpy as np
from pydantic import BaseModel
from sqlalchemy import asc, desc, select
from sqlalchemy.orm import Session

from autoflow.storage.tidb.constants import DistanceMetric, VectorDataType
from autoflow.storage.tidb.utils import build_filter_clauses, check_vector_column

if TYPE_CHECKING:
    from autoflow.storage.tidb.table import Table


class Query(BaseModel):
    vector_column: Optional[str] = None
    vector: Union[List[float], List[List[float]]]
    filter: Optional[Dict[str, Any]] = None
    distance_metric: str = "L2"
    distance_range: Optional[Tuple[float, float]] = None
    nprobes: int = 10
    top_k: Optional[int] = None
    use_index: bool = True


class QueryType(str, Enum):
    VECTOR_SEARCH = "vector_search"
    FULLTEXT_SEARCH = "fulltext_search"
    HYBRID_SEARCH = "hybrid_search"


class TiDBQuery(ABC):
    def __init__(self, table: "Table"):
        self._table = table
        self._limit = None
        self._offset = 0
        self._columns = None
        self._where = None
        self._prefilter = True
        self._with_row_id = False
        self._vector = None
        self._text = None
        self._ef = None
        self._use_index = True

    @classmethod
    def create(
        cls,
        table: "Table",
        query: Optional[Union[np.ndarray, str, Tuple]],
        query_type: QueryType,
    ) -> "TiDBQuery":
        if query_type == QueryType.VECTOR_SEARCH:
            return TiDBVectorQuery(
                table=table,
                query=query,
            )
        elif query_type == QueryType.FULLTEXT_SEARCH:
            raise NotImplementedError
        elif query_type == QueryType.HYBRID_SEARCH:
            raise NotImplementedError
        else:
            raise NotImplementedError

    def limit(self, limit: Union[int, None]) -> "TiDBQuery":
        if limit is None or limit <= 0:
            if isinstance(self, TiDBVectorQuery):
                raise ValueError("Limit is required for ANN/KNN queries")
            else:
                self._limit = None
        else:
            self._limit = limit
        return self

    def execute(self):
        return self._execute()

    @abstractmethod
    def _execute(self):
        pass

    def to_rows(self) -> Sequence[Any]:
        return self.execute()


class TiDBVectorQuery(TiDBQuery):
    def __init__(self, table: "Table", query: VectorDataType):
        super().__init__(table)
        if self._limit is None:
            self._limit = 10
        self._query = query
        self._distance_metric = DistanceMetric.COSINE
        self._num_candidate = 20
        self._vector_column = table.vector_column
        self._filters = None

    def vector_column(self, column_name: str):
        self._vector_column = check_vector_column(self._columns, column_name)
        return self

    def distance_metric(self, metric: DistanceMetric) -> "TiDBVectorQuery":
        self._distance_metric = metric
        return self

    def num_candidate(self, num_candidate: int) -> "TiDBVectorQuery":
        self._num_candidate = num_candidate
        return self

    def filter(self, filters: Optional[Dict[str, Any]] = None) -> "TiDBVectorQuery":
        self._filters = filters
        return self

    def limit(self, k: int) -> "TiDBVectorQuery":
        self._limit = k
        return self

    def _execute(self) -> Sequence[Any]:
        num_candidate = self._num_candidate if self._num_candidate else self._limit * 10

        if self._vector_column is None:
            if len(self._table.vector_columns) == 0:
                raise ValueError(
                    "no vector column found in the table, vector search cannot be executed"
                )
            elif len(self._table.vector_columns) >= 1:
                raise ValueError(
                    "more than two vector columns in the table, need to be specified one through .vector_column()"
                )
            else:
                vector_column = self._table.vector_columns[0]
        else:
            vector_column = self._vector_column

        # Auto embedding
        if isinstance(self._query, str):
            if vector_column.name not in self._table.vector_field_configs:
                raise ValueError()

            config = self._table.vector_field_configs[vector_column.name]
            self._query = config["embed_fn"].get_query_embedding(self._query)

        # Distance metric.
        distance_label = "_distance"
        if self._distance_metric == DistanceMetric.L2:
            distance_column = vector_column.l2_distance(self._query).label(
                distance_label
            )
        else:
            distance_column = vector_column.cosine_distance(self._query).label(
                distance_label
            )

        # Inner query for ANN search
        db_engine = self._table.db_engine
        table_model = self._table.table_model
        columns = table_model.__table__.c
        subquery = (
            select(columns, distance_column)
            .order_by(asc(distance_label))
            .limit(num_candidate)
            .subquery("candidates")
        )

        # Main query with metadata filters
        query = select(
            subquery.c,
            (1 - subquery.c[distance_label]).label("similarity_score"),
        )

        if self._filters is not None:
            filter_clauses = build_filter_clauses(
                self._filters, subquery.c, table_model
            )
            query = query.filter(*filter_clauses)

        query = query.order_by(desc("similarity_score")).limit(self._limit)

        sql = query.compile(dialect=db_engine.dialect)
        logging.info(sql)

        with Session(db_engine) as session:
            return session.execute(query).all()

    def to_pydantic(self) -> List[BaseModel]:
        rows = self.execute()
        results = []
        for row in rows:
            model = self._table.table_model.model_validate(row._mapping)
            if row._mapping["similarity_score"] is not None:
                model.similarity_score = row._mapping["similarity_score"]
                model.score = row._mapping["similarity_score"]
            results.append(model)
        return results



import uuid
from abc import abstractmethod
from typing import Generator, Generic, TypeVar, Optional
from pydantic import BaseModel, Field

from autoflow.models import DBDocument
from autoflow.schema import BaseComponent

C = TypeVar("C", bound=BaseModel)


class DataSource(BaseComponent, Generic[C]):
    id: uuid.UUID = Field(default_factory=uuid.uuid4)
    config: Optional[C] = Field(default=None)

    def __init__(self, config: dict):
        config = self.validate_config(config)
        super().__init__(config=config)

    @abstractmethod
    def validate_config(self, config: dict) -> C:
        raise NotImplementedError()

    @abstractmethod
    def load_documents(self) -> Generator[DBDocument, None, None]:
        raise NotImplementedError



import logging
from typing import List

import dspy
from dspy import Predict

from autoflow.indices.knowledge_graph.schema import (
    AIKnowledgeGraph,
    EntityCovariateInput,
    EntityCovariateOutput,
)

logger = logging.getLogger(__name__)


class ExtractGraphTriplet(dspy.Signature):
    """Carefully analyze the provided text from database documentation and community blogs to thoroughly identify all entities related to database technologies, including both general concepts and specific details.

    Follow these Step-by-Step Analysis:

    1. Extract Meaningful Entities:
      - Identify all significant nouns, proper nouns, and technical terminologies that represent database-related concepts, objects, components, features, issues, key steps, execute order, user case, locations, versions, or any substantial entities.
      - Ensure that you capture entities across different levels of detail, from high-level overviews to specific technical specifications, to create a comprehensive representation of the subject matter.
      - Choose names for entities that are specific enough to indicate their meaning without additional context, avoiding overly generic terms.
      - Consolidate similar entities to avoid redundancy, ensuring each represents a distinct concept at appropriate granularity levels.

    2. Extract Metadata to claim the entities:
      - Carefully review the provided text, focusing on identifying detailed covariates associated with each entity.
      - Extract and link the covariates (which is a comprehensive json TREE, the first field is always: "topic") to their respective entities.
      - Ensure all extracted covariates is clearly connected to the correct entity for accuracy and comprehensive understanding.
      - Ensure that all extracted covariates are factual and verifiable within the text itself, without relying on external knowledge or assumptions.
      - Collectively, the covariates should provide a thorough and precise summary of the entity's characteristics as described in the source material.

    3. Establish Relationships:
      - Carefully examine the text to identify all relationships between clearly-related entities, ensuring each relationship is correctly captured with accurate details about the interactions.
      - Analyze the context and interactions between the identified entities to determine how they are interconnected, focusing on actions, associations, dependencies, or similarities.
      - Clearly define the relationships, ensuring accurate directionality that reflects the logical or functional dependencies among entities. \
         This means identifying which entity is the source, which is the target, and what the nature of their relationship is (e.g., $source_entity depends on $target_entity for $relationship).

    Some key points to consider:
      - Please endeavor to extract all meaningful entities and relationships from the text, avoid subsequent additional gleanings.

    Objective: Produce a detailed and comprehensive knowledge graph that captures the full spectrum of entities mentioned in the text, along with their interrelations, reflecting both broad concepts and intricate details specific to the database domain.

    Please only response in JSON format.
    """

    text = dspy.InputField(
        desc="a paragraph of text to extract entities and relationships to form a knowledge graph"
    )
    knowledge: AIKnowledgeGraph = dspy.OutputField(
        desc="Graph representation of the knowledge extracted from the text."
    )


class ExtractCovariate(dspy.Signature):
    """Please carefully review the provided text and entities list which are already identified in the text. Focusing on identifying detailed covariates associated with each entities provided.
    Extract and link the covariates (which is a comprehensive json TREE, the first field is always: "topic") to their respective entities.
    Ensure all extracted covariates is clearly connected to the correct entity for accuracy and comprehensive understanding.
    Ensure that all extracted covariates are factual and verifiable within the text itself, without relying on external knowledge or assumptions.
    Collectively, the covariates should provide a thorough and precise summary of the entity's characteristics as described in the source material.

    Please only response in JSON format.
    """

    text = dspy.InputField(
        desc="a paragraph of text to extract covariates to claim the entities."
    )

    entities: List[EntityCovariateInput] = dspy.InputField(
        desc="List of entities identified in the text."
    )
    covariates: List[EntityCovariateOutput] = dspy.OutputField(
        desc="Graph representation of the knowledge extracted from the text."
    )


class KnowledgeGraphExtractor(dspy.Module):
    def __init__(self, dspy_lm: dspy.LM):
        super().__init__()
        self.dspy_lm = dspy_lm
        self.prog_graph = Predict(ExtractGraphTriplet)
        self.prog_covariates = Predict(ExtractCovariate)

    def forward(self, text: str) -> AIKnowledgeGraph:
        with dspy.settings.context(lm=self.dspy_lm):
            pred_graph = self.prog_graph(
                text=text,
            )

            # Extract the covariates.
            entities_for_covariates = [
                EntityCovariateInput(
                    name=entity.name,
                    description=entity.description,
                )
                for entity in pred_graph.knowledge.entities
            ]

            pred_covariates = self.prog_covariates(
                text=text,
                entities=entities_for_covariates,
            )

            # Replace the entities with the covariates.
            for entity in pred_graph.knowledge.entities:
                for covariate in pred_covariates.covariates:
                    if entity.name == covariate.name:
                        entity.metadata = covariate.covariates

            return pred_graph



from abc import abstractmethod, ABC
from typing import Tuple, Generic, List

from pydantic import BaseModel

from ..base import KnowledgeGraphStore, E, R
from ...schema import QueryBundle


class EntityWithScore(BaseModel, Generic[E]):
    entity: E
    score: float

    @property
    def id(self) -> int:
        return self.entity.id

    def __hash__(self) -> int:
        return hash(self.entity.id)


class RelationshipWithScore(Generic[R]):
    relationship: R
    score: float

    def __init__(self, relationship: R, score: float):
        self.relationship = relationship
        self.score = score

    @property
    def id(self) -> int:
        return self.relationship.id

    def __hash__(self) -> int:
        return hash(self.relationship.id)


class KnowledgeGraphRetriever(ABC, Generic[E, R]):
    def __init__(self, knowledge_graph_store: "KnowledgeGraphStore"):
        self._kg_store = knowledge_graph_store

    @abstractmethod
    def search(
        self,
        query: QueryBundle,
        depth: int = 2,
        meta_filters: dict = None,
    ) -> Tuple[List[RelationshipWithScore[R]], List[E]]:
        raise NotImplementedError



r"""UUID draft version objects (universally unique identifiers).
This module provides the functions uuid6() and uuid7() for
generating version 6 and 7 UUIDs as specified in
https://github.com/uuid6/uuid6-ietf-draft.

Repo: https://github.com/oittaa/uuid6-python

Copy from: https://github.com/jonra1993/fastapi-alembic-sqlmodel-async/blob/main/backend/app/app/utils/uuid6.py
"""

import secrets
import time
import uuid


class UUID(uuid.UUID):
    r"""UUID draft version objects"""

    def __init__(
        self,
        hex: str = None,
        bytes: bytes = None,
        bytes_le: bytes = None,
        fields: tuple[int, int, int, int, int, int] = None,
        int: int = None,
        version: int = None,
        *,
        is_safe=uuid.SafeUUID.unknown,
    ) -> None:
        r"""Create a UUID."""

        if int is None or [hex, bytes, bytes_le, fields].count(None) != 4:
            super().__init__(
                hex=hex,
                bytes=bytes,
                bytes_le=bytes_le,
                fields=fields,
                int=int,
                version=version,
                is_safe=is_safe,
            )
        if not 0 <= int < 1 << 128:
            raise ValueError("int is out of range (need a 128-bit value)")
        if version is not None:
            if not 6 <= version <= 7:
                raise ValueError("illegal version number")
            # Set the variant to RFC 4122.
            int &= ~(0xC000 << 48)
            int |= 0x8000 << 48
            # Set the version number.
            int &= ~(0xF000 << 64)
            int |= version << 76
        super().__init__(int=int, is_safe=is_safe)

    @property
    def subsec(self) -> int:
        return ((self.int >> 64) & 0x0FFF) << 8 | ((self.int >> 54) & 0xFF)

    @property
    def time(self) -> int:
        if self.version == 6:
            return (
                (self.time_low << 28)
                | (self.time_mid << 12)
                | (self.time_hi_version & 0x0FFF)
            )
        if self.version == 7:
            return (self.int >> 80) * 10**6 + _subsec_decode(self.subsec)
        return super().time


def _subsec_decode(value: int) -> int:
    return -(-value * 10**6 // 2**20)


def _subsec_encode(value: int) -> int:
    return value * 2**20 // 10**6


_last_v6_timestamp = None
_last_v7_timestamp = None


def uuid6(clock_seq: int = None) -> UUID:
    r"""UUID version 6 is a field-compatible version of UUIDv1, reordered for
    improved DB locality.  It is expected that UUIDv6 will primarily be
    used in contexts where there are existing v1 UUIDs.  Systems that do
    not involve legacy UUIDv1 SHOULD consider using UUIDv7 instead.
    If 'clock_seq' is given, it is used as the sequence number;
    otherwise a random 14-bit sequence number is chosen."""

    global _last_v6_timestamp

    nanoseconds = time.time_ns()
    # 0x01b21dd213814000 is the number of 100-ns intervals between the
    # UUID epoch 1582-10-15 00:00:00 and the Unix epoch 1970-01-01 00:00:00.
    timestamp = nanoseconds // 100 + 0x01B21DD213814000
    if _last_v6_timestamp is not None and timestamp <= _last_v6_timestamp:
        timestamp = _last_v6_timestamp + 1
    _last_v6_timestamp = timestamp
    if clock_seq is None:
        clock_seq = secrets.randbits(14)  # instead of stable storage
    node = secrets.randbits(48)
    time_high_and_time_mid = (timestamp >> 12) & 0xFFFFFFFFFFFF
    time_low_and_version = timestamp & 0x0FFF
    uuid_int = time_high_and_time_mid << 80
    uuid_int |= time_low_and_version << 64
    uuid_int |= (clock_seq & 0x3FFF) << 48
    uuid_int |= node
    return UUID(int=uuid_int, version=6)


def uuid7() -> UUID:
    r"""UUID version 7 features a time-ordered value field derived from the
    widely implemented and well known Unix Epoch timestamp source, the
    number of milliseconds seconds since midnight 1 Jan 1970 UTC, leap
    seconds excluded.  As well as improved entropy characteristics over
    versions 1 or 6.
    Implementations SHOULD utilize UUID version 7 over UUID version 1 and
    6 if possible."""

    global _last_v7_timestamp

    nanoseconds = time.time_ns()
    if _last_v7_timestamp is not None and nanoseconds <= _last_v7_timestamp:
        nanoseconds = _last_v7_timestamp + 1
    _last_v7_timestamp = nanoseconds
    timestamp_ms, timestamp_ns = divmod(nanoseconds, 10**6)
    subsec = _subsec_encode(timestamp_ns)
    subsec_a = subsec >> 8
    subsec_b = subsec & 0xFF
    rand = secrets.randbits(54)
    uuid_int = (timestamp_ms & 0xFFFFFFFFFFFF) << 80
    uuid_int |= subsec_a << 64
    uuid_int |= subsec_b << 54
    uuid_int |= rand
    return UUID(int=uuid_int, version=7)



import logging
from typing import Any, List

import llama_index.core.instrumentation as instrument
from sqlmodel import SQLModel
from autoflow.storage import DocumentStore
from autoflow.schema import BaseComponent

logger = logging.getLogger(__name__)


dispatcher = instrument.get_dispatcher(__name__)


class VectorSearchIndex(BaseComponent):
    def __init__(
        self,
        doc_store: DocumentStore,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self._doc_store = doc_store

    def build_index_for_chunks(self, chunks: List[SQLModel]):
        if len(chunks) == 0:
            return chunks

        self._doc_store.add_doc_chunks(chunks)



import uuid
from typing import List, Dict

from sqlmodel import Field
from tidb_vector.sqlalchemy import VectorType
from autoflow.storage.tidb.base import TiDBModel


class Memory(TiDBModel, table=True):
    __tablename__ = "agent_memory"
    id: uuid.UUID = Field(None, primary_key=True)
    vector: List[float] = Field(None, sa_type=VectorType)
    payload: Dict = Field(None)



import enum
from typing import Optional, Any, List, Dict, Type

from sqlalchemy.orm.decl_api import RegistryType
from sqlmodel import (
    Field,
    Column,
    JSON,
    Text,
)
from sqlalchemy import Index
from sqlmodel.main import default_registry, SQLModel
from tidb_vector.sqlalchemy import VectorType

from autoflow.utils.sql_model import PatchSQLModel


class EntityType(str, enum.Enum):
    original = "original"
    synopsis = "synopsis"

    def __str__(self):
        return self.value


def get_entity_model(
    table_name: str,
    vector_dimension: int,
    registry: RegistryType = default_registry,
) -> Type[SQLModel]:
    class Entity(PatchSQLModel, table=True, registry=registry):
        __tablename__ = table_name
        __table_args__ = (
            Index("idx_entity_type", "entity_type"),
            Index("idx_entity_name", "name"),
            {"extend_existing": True},
        )

        id: Optional[int] = Field(default=None, primary_key=True)
        entity_type: EntityType = EntityType.original
        name: str = Field(max_length=512)
        description: str = Field(sa_column=Column(Text))
        description_vec: list[float] = Field(
            default=None, sa_column=Column(VectorType(vector_dimension))
        )
        meta: List | Dict = Field(default={}, sa_column=Column(JSON))
        meta_vec: list[float] = Field(
            default=None, sa_column=Column(VectorType(vector_dimension))
        )
        synopsis_info: List | Dict | None = Field(default=None, sa_column=Column(JSON))

        def __hash__(self):
            return hash(self.id)

        def __eq__(self, other: Any) -> bool:
            return self.id == other.id

        def screenshot(self):
            return self.model_dump(exclude={"description_vec", "meta_vec"})

    return Entity



import enum
import typing
import numpy
from sqlalchemy.orm import DeclarativeMeta
from sqlmodel.main import SQLModelMetaclass

# TiDB Vector has a limitation on the dimension length
MAX_DIM = 16000
MIN_DIM = 1

# Filter operators:

AND, OR, IN, NIN, GT, GTE, LT, LTE, EQ, NE = (
    "$and",
    "$or",
    "$in",
    "$nin",
    "$gt",
    "$gte",
    "$lt",
    "$lte",
    "$eq",
    "$ne",
)

COMPARE_OPERATOR = [IN, NIN, GT, GTE, LT, LTE, EQ, NE]

TableModel = typing.Union[SQLModelMetaclass, DeclarativeMeta]

VectorDataType = typing.Union[numpy.ndarray, typing.List[float]]


class DistanceMetric(enum.Enum):
    """
    An enumeration representing different types of distance metrics.

    - `DistanceMetric.L2`: L2 (Euclidean) distance metric.
    - `DistanceMetric.COSINE`: Cosine distance metric.
    """

    L2 = "L2"
    COSINE = "COSINE"

    def to_sql_func(self):
        """
        Converts the DistanceMetric to its corresponding SQL function name.

        Returns:
            str: The SQL function name.

        Raises:
            ValueError: If the DistanceMetric enum member is not supported.
        """
        if self == DistanceMetric.L2:
            return "VEC_L2_DISTANCE"
        elif self == DistanceMetric.COSINE:
            return "VEC_COSINE_DISTANCE"
        else:
            raise ValueError("unsupported distance metric")



# flake8: noqa
from .document import Document as DBDocument
from .chunk import get_chunk_model
from .entity import get_entity_model, EntityType
from .relationship import get_relationship_model

__all__ = [
    "DBDocument",
    "get_chunk_model",
    "get_entity_model",
    "EntityType",
    "get_relationship_model",
]



import logging
from typing import Generator
from urllib.parse import urlparse, urljoin

import requests
from pydantic import BaseModel

from autoflow.datasources.base import DataSource
from autoflow.datasources.web_base import load_web_documents
from autoflow.models import DBDocument

logger = logging.getLogger(__name__)


class WebSitemapDataSourceConfig(BaseModel):
    url: str


def _ensure_absolute_url(source_url: str, maybe_relative_url: str) -> str:
    if not urlparse(maybe_relative_url).netloc:
        return urljoin(source_url, maybe_relative_url)
    return maybe_relative_url


def extract_urls_from_sitemap(sitemap_url: str) -> list[str]:
    from bs4 import BeautifulSoup

    response = requests.get(sitemap_url)
    response.raise_for_status()

    soup = BeautifulSoup(response.content, "html.parser")
    result = [
        _ensure_absolute_url(sitemap_url, loc_tag.text)
        for loc_tag in soup.find_all("loc")
    ]
    if not result:
        raise ValueError(f"No URLs found in sitemap {sitemap_url}")
    return result


class WebSitemapDataSource(DataSource):
    def validate_config(self, config: dict):
        return WebSitemapDataSourceConfig.model_validate(config)

    def load_documents(self) -> Generator[DBDocument, None, None]:
        sitemap_url = self.config["url"]
        urls = extract_urls_from_sitemap(sitemap_url)
        logger.info(f"Found {len(urls)} URLs in sitemap {sitemap_url}")
        for doc in load_web_documents(urls):
            doc.data_source_id = self.id
            yield doc



from typing import List

from .litellm import LiteLLMReranker
from ...storage.tidb.base import TiDBModel
from ...storage.tidb.rerank_fn import BaseRerankFunction

RerankerModel = LiteLLMReranker


class RerankFunction(BaseRerankFunction):
    def __init__(self, model: str):
        self._model = model

    def rerank(
        self, items: List[TiDBModel], query_str: str, top_n: int = 2
    ) -> List[TiDBModel]:
        raise NotImplementedError("Rerank function is not implemented.")
        # reranker_model = RerankerModel(model=self._model, top_n=top_n)
        # nodes = [
        #     NodeWithScore(
        #         node=TextNode(
        #             text=item.sou
        #         )
        #     )
        #     for item in items
        # ]
        # reranker_model.postprocess_nodes(nodes=)
        # return items


__all__ = ["RerankerModel"]



import logging
from typing import List, Optional

import sqlalchemy
from pydantic import PrivateAttr
from sqlalchemy.engine import Engine, create_engine
from sqlalchemy.orm import Session

from autoflow.storage.tidb.table import Table, TableModel
from autoflow.storage.tidb.utils import build_tidb_dsn

logger = logging.getLogger(__name__)


class TiDBClient:
    _db_engine: Engine = PrivateAttr()

    def __init__(self, db_engine: Engine):
        self._db_engine = db_engine
        self._inspector = sqlalchemy.inspect(self._db_engine)

    @classmethod
    def connect(
        cls,
        database_url: Optional[str] = None,
        *,
        host: Optional[str] = "localhost",
        port: Optional[int] = 4000,
        username: Optional[str] = "root",
        password: Optional[str] = "",
        database: Optional[str] = "test",
        enable_ssl: Optional[bool] = None,
        **kwargs,
    ) -> "TiDBClient":
        if database_url is None:
            database_url = str(
                build_tidb_dsn(
                    host=host,
                    port=port,
                    username=username,
                    password=password,
                    database=database,
                    enable_ssl=enable_ssl,
                )
            )

        db_engine = create_engine(database_url, **kwargs)
        return cls(db_engine)

    # Notice: Since the Vector type is not in the type support list of mysql dialect, using the reflection API will cause an error.
    # https://github.com/sqlalchemy/sqlalchemy/blob/d6f11d9030b325d5afabf87869a6e3542edda54b/lib/sqlalchemy/dialects/mysql/base.py#L1199
    # def _load_table_metadata(self, table_names: Optional[List[str]] = None):
    #     if not table_names:
    #         Base.metadata.reflect(bind=self._db_engine)
    #     else:
    #         Base.metadata.reflect(bind=self._db_engine, only=table_names, extend_existing=True)

    @property
    def db_engine(self) -> Engine:
        return self._db_engine

    def create_table(
        self,
        *,
        schema: Optional[TableModel] = None,
    ) -> Table:
        table = Table(schema=schema, db_engine=self._db_engine)
        return table

    def open_table(self, schema: TableModel) -> Table:
        return Table(
            schema=schema,
            db_engine=self._db_engine,
        )

    def table_names(self) -> List[str]:
        return self._inspector.get_table_names()

    def has_table(self, table_name: str) -> bool:
        return self._inspector.has_table(table_name)

    def drop_table(self, table_name: str):
        return self.execute(f"DROP TABLE IF EXISTS {table_name}")

    def execute(self, sql: str, params: Optional[dict] = None) -> dict:
        """
        Execute an arbitrary SQL command and return execution status and result.

        This method can handle both DML (Data Manipulation Language) commands such as INSERT, UPDATE, DELETE,
        and DQL (Data Query Language) commands like SELECT. It returns a structured dictionary indicating
        the execution success status, result (for SELECT queries or affected rows count for DML), and any
        error message if the execution failed.

        Args:
            sql (str): The SQL command to execute.
            params (Optional[dict]): Parameters to bind to the SQL command, if any.

        Returns:
            dict: A dictionary containing 'success': boolean indicating if the execution was successful,
                'result': fetched results for SELECT or affected rows count for other statements,
                and 'error': error message if execution failed.

        Examples:
            - Creating a table:
            execute("CREATE TABLE users (id INT, username VARCHAR(50), email VARCHAR(50))")
            This would return: {'success': True, 'result': 0, 'error': None}

            - Executing a SELECT query:
            execute("SELECT * FROM users WHERE username = :username", {"username": "john_doe"})
            This would return: {'success': True, 'result': [(user data)], 'error': None}

            - Inserting data into a table:
            execute(
                "INSERT INTO users (username, email) VALUES (:username, :email)",
                {"username": "new_user", "email": "new_user@example.com"}
            )
            This would return: {'success': True, 'result': 1, 'error': None} if one row was affected.

            - Handling an error (e.g., table does not exist):
            execute("SELECT * FROM non_existing_table")
            This might return: {'success': False, 'result': None, 'error': '(Error message)'}
        """
        try:
            with Session(self._db_engine) as session, session.begin():
                result = session.execute(sqlalchemy.text(sql), params)
                session.commit()  # Ensure changes are committed for non-SELECT statements.
                if sql.strip().lower().startswith("select"):
                    return {"success": True, "result": result.fetchall(), "error": None}
                else:
                    return {"success": True, "result": result.rowcount, "error": None}
        except Exception as e:
            # Log the error or handle it as needed
            logger.error(f"SQL execution error: {str(e)}")
            return {"success": False, "result": None, "error": str(e)}

    def disconnect(self) -> None:
        self._db_engine.dispose()



import logging
from datetime import datetime, UTC
from typing import Generator

from autoflow.datasources.mime_types import SupportedMimeTypes
from autoflow.models import DBDocument


logger = logging.getLogger(__name__)


IGNORE_TAGS = [
    "noscript",
    "title",
    "script",
    "style",
    "meta",
    "head",
    "header",
    "footer",
    "nav",
    "symbol",
    "aside",
]

IGNORE_CLASSES = ["header", "footer", "sidebar"]


def load_web_documents(urls: list[str]) -> Generator[DBDocument, None, None]:
    from playwright.sync_api import sync_playwright
    from bs4 import BeautifulSoup
    from markdownify import MarkdownConverter

    visited = set()
    with sync_playwright() as p:
        browser = p.chromium.launch(headless=True)
        for url in urls:
            page = browser.new_page()
            response = page.goto(url)
            final_url = page.url
            if final_url in visited:
                continue

            if response.status >= 400:
                logger.error(
                    f"Failed to load page: {url}, response status: {response.status()}, skipping"
                )
                continue
            soup = BeautifulSoup(page.content(), "html.parser")
            for t in IGNORE_TAGS:
                for tag in soup.find_all(t):
                    tag.extract()
            for c in IGNORE_CLASSES:
                for tag in soup.find_all(class_=c):
                    tag.extract()
            content = MarkdownConverter().convert_soup(soup)
            title = page.title()
            visited.add(final_url)
            document = DBDocument(
                name=title,
                hash=hash(content),
                content=content,
                mime_type=SupportedMimeTypes.PLAIN_TXT,
                source_uri=final_url,
                last_modified_at=datetime.now(UTC),
            )
            yield document
        browser.close()



import logging
import mimetypes
from pathlib import Path
from typing import Generator, List, Optional

from llama_index.core.readers.base import BaseReader
from pydantic import BaseModel

from .base import DataSource, C
from .mime_types import SupportedMimeTypes
from autoflow.models import DBDocument

logger = logging.getLogger(__name__)


class FileConfig(BaseModel):
    path: str


class FileDataSourceConfigConfig(BaseModel):
    files: List[FileConfig]


def get_file_reader_by_mime_type(mime_type: str) -> Optional[BaseReader]:
    if mime_type == SupportedMimeTypes.PDF:
        from llama_index.readers.file import PDFReader

        return PDFReader(return_full_document=True)
    elif mime_type == SupportedMimeTypes.DOCX:
        from llama_index.readers.file import DocxReader

        return DocxReader()
    elif mime_type == SupportedMimeTypes.PPTX:
        from llama_index.readers.file import PptxReader

        return PptxReader()
    elif mime_type == SupportedMimeTypes.XLSX:
        from llama_index.readers.file import PandasExcelReader

        return PandasExcelReader()
    elif mime_type == SupportedMimeTypes.CSV:
        from llama_index.readers.file import CSVReader

        return CSVReader()
    elif mime_type == SupportedMimeTypes.MARKDOWN:
        from llama_index.readers.file import MarkdownReader

        return MarkdownReader()
    else:
        return None


class FileDataSource(DataSource[FileDataSourceConfigConfig]):
    def validate_config(self, config: dict) -> C:
        return FileDataSourceConfigConfig.model_validate(config)

    def load_documents(self) -> Generator[DBDocument, None, None]:
        for fc in self.config.files:
            filepath = Path(fc.path)
            filename = filepath.name
            mime_type = mimetypes.guess_type(fc.path)[0]
            if mime_type is None:
                if fc.path.endswith(".md"):
                    mime_type = SupportedMimeTypes.MARKDOWN
            reader = get_file_reader_by_mime_type(mime_type)
            docs = reader.load_data(filepath)
            for doc in docs:
                document = DBDocument(
                    name=filename,
                    hash=hash(doc.text),
                    mime_type=SupportedMimeTypes.MARKDOWN
                    if mime_type == SupportedMimeTypes.MARKDOWN
                    else SupportedMimeTypes.PLAIN_TXT,
                    content=doc.text,
                )
                yield document



from .client import TiDBClient
from .base import default_registry, Base
from .table import Table
from .errors import EmbeddingColumnMismatchError
from .constants import DistanceMetric
from .base import TiDBModel, Field

__all__ = [
    "TiDBClient",
    "Table",
    "default_registry",
    "Base",
    "EmbeddingColumnMismatchError",
    "DistanceMetric",
    "TiDBModel",
    "Field",
]
